# from fire import Fire
from datasets import load_dataset, get_dataset_config_names, DatasetDict
from datasets import Image as DSImage
from tqdm.auto import tqdm
import pandas as pd
from PIL import Image
from pathlib import Path
import ast
import re
from multiprocessing import Pool
from functools import partial
import json

HF_DATASET = "MMMU/MMMU"


def _download_hf_dataset(
    configs: list[str] | None, n_proc: int | None = None
) -> dict[str, DatasetDict]:
    cfgs = get_dataset_config_names(HF_DATASET) if configs is None else configs
    ds = {}
    for c in (pbar := tqdm(cfgs, desc="Downloading HF Dataset")):
        pbar.set_postfix_str(f"Config: {c}")
        hf_ds = load_dataset(HF_DATASET, c, num_proc=n_proc)
        ds[c] = hf_ds
    return ds


def _transform_to_dataframe(hf_datasets: dict[str, DatasetDict]) -> pd.DataFrame:
    full_df = None
    for cfg, hf_ds in (
        pbar := tqdm(
            hf_datasets.items(),
            total=len(hf_datasets),
            desc="Transforming HF Dataset to Pandas",
            position=0,
        )
    ):
        pbar.set_postfix_str(f"Config: {cfg}")
        df = pd.concat(split.to_pandas() for split in hf_ds.values())
        df.insert(1, "cfg", cfg)
        if full_df is None:
            full_df = df
        else:
            full_df = pd.concat([full_df, df])

    return full_df


def _store_images_for_sample(sample, image_out_p: Path) -> dict:
    for i in range(1, 8):
        img_fn = image_out_p / f"{sample['id']}_image_{i}.png"
        if not img_fn.exists():
            try:
                img = sample[f"image_{i}"]
                if img is not None and not isinstance(img, Image.Image):
                    img = DSImage().decode_example(sample[f"image_{i}"])
                elif isinstance(img, Image.Image):
                    pass
                else:
                    img = None

                if img is None:
                    sample[f"image_{i}"] = None
                    continue

                img.save(img_fn)
                sample[f"image_{i}"] = img_fn.name
            except Exception as e:
                print(f"Error saving image {img_fn}: {e}")
                sample[f"image_{i}"] = None
        else:
            sample[f"image_{i}"] = img_fn.name

    return sample


def _store_images_for_dataset(
    cfg: str, ds: DatasetDict, image_out_p: Path
) -> tuple[str, DatasetDict]:
    for split in ds.keys():
        ds[split] = ds[split].map(partial(_store_images_for_sample, image_out_p=image_out_p))
    return cfg, ds


def _create_json_files_for_trident_dataspec(
    full_df: pd.DataFrame, json_out_p: Path
) -> list[Path]:
    full_df.options = full_df.options.apply(ast.literal_eval)

    for split in ["dev", "test", "validation"]:
        split_df = full_df[full_df["id"].str.startswith(f"{split}_")]

        samples = []

        for ridx, row in split_df.iterrows():
            text_label = row["answer"]

            question = row["question"]
            if row["question_type"] == "multiple-choice":
                options: str = "\n".join(
                    f"{list('ABCDEFGHI')[i]}. {o}" for i, o in enumerate(row["options"])
                )

                question += (
                    "\nThere are several options:\n"
                    f"{options}"
                    "\nAnswer with the option's letter from the given choices directly."
                )
            else:
                question += (
                    "\nProvide your answer in the form of a single word or phrase."
                )

            # get all image tags in the right order
            pattern = r"<image \d>"
            image_tags = re.findall(pattern, question)
            image_tag_ids = [int(tag.split(" ")[1][:-1]) for tag in image_tags]
            image_id = [row[f"image_{i}"]["path"] for i in image_tag_ids]

            for i in range(1, 8):
                question = question.replace(f"<image {i}>", "<image_placeholder>")

            assert (
                len(image_id) == question.count("<image_placeholder>")
            ), f"Mismatch in sample {ridx} {row['id']} \n {question=} {image_id=}, {image_tags=}"

            sample = {
                "context": question,
                "image_id": image_id,
                "text_label": text_label,
                "label": "",
            }

            # if row["explanation"] is not None and "<image" in row["explanation"]:

            # assert len(image_id) == question.count(
            #     "<image_placeholder>"
            # ), f"Mismatch in sample {ridx} {row['id']}"

            samples.append(sample)

        ofn = json_out_p / f"mmmu_en_{split}.json"
        with open(ofn, "w") as f:
            json.dump(samples, f)
        print(f"Saved MMMU {split} dataspec with {len(samples)} to {ofn}")


def run(
    json_out_p: str | Path = ".",
    image_out_p: str | Path = "./images",
    hf_cache: str | Path = "~/.cache/huggingface/datasets",
    n_proc: int = 30,
    configs: tuple[str] | None = None,
    skip_images: bool = False,
):
    json_out_p = Path(json_out_p)
    image_out_p = Path(image_out_p)
    hf_cache = Path(hf_cache).expanduser()

    json_out_p.mkdir(parents=True, exist_ok=True)
    image_out_p.mkdir(parents=True, exist_ok=True)
    hf_cache.mkdir(parents=True, exist_ok=True)

    if not skip_images:
        hf_ds = _download_hf_dataset(configs, n_proc)

        if n_proc == 1:
            for cfg, ds in hf_ds.items():
                hf_ds[cfg] = _store_images_for_dataset(ds)
        else:
            n_proc = min(n_proc, len(hf_ds))
            with Pool(n_proc) as p:
                results = p.starmap(
                    _store_images_for_dataset,
                    [(cfg, ds, image_out_p) for cfg, ds in hf_ds.items()],
                )
            hf_ds = {cfg: ds for cfg, ds in results}
        print(f"Saved images to {image_out_p}")

        full_df = _transform_to_dataframe(hf_ds)
        ofn = json_out_p / "full_mmmu.json"
        full_df.to_json(ofn, orient="records", lines=True)
        print(f"Saved full MMMU dataset to {ofn}")

    else:
        full_df = pd.read_json(json_out_p / "full_mmmu.json", lines=True)

    _create_json_files_for_trident_dataspec(full_df, json_out_p)


if __name__ == "__main__":
    # Fire(run)
    run(image_out_p="/media/gregor/DATA/projects/wuerzburg/iglue/datasets/mmmu/images")